@@ -35,6 +35,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <dirent.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 
@@ -97,6 +98,15 @@ static char *backlight_interfaces[] = {
 
 static int backlight_index;
 
+enum lid_status {
+    LID_UNKNOWN = -1,
+    LID_OPEN,
+    LID_CLOSE,
+};
+
+#define ACPI_BUTTON "/proc/acpi/button/"
+#define ACPI_LID "/proc/acpi/button/lid/"
+
 static Bool
 i830_kernel_backlight_available(xf86OutputPtr output)
 {
@@ -376,6 +386,107 @@ out_err:
     return 0;
 }
 
+/**
+ *  Get lid state from ACPI button driver
+ */
+static int
+i830_lvds_acpi_lid_open(xf86OutputPtr output)
+{
+    ScrnInfoPtr pScrn = output->scrn;
+    I830Ptr pI830 = I830PTR(pScrn);
+    int fd;
+    DIR *button_dir;
+    DIR *lid_dir;
+    struct dirent *lid_dent;
+    char *state_name;
+    char state[64];
+    enum lid_status ret = LID_UNKNOWN;
+
+    if (pI830->quirk_flag & QUIRK_BROKEN_ACPI_LID)
+	goto out;
+
+    button_dir = opendir(ACPI_BUTTON);
+    /* If acpi button driver is not loaded, bypass ACPI check method */
+    if (button_dir == NULL)
+	goto out;
+    closedir(button_dir);
+
+    lid_dir = opendir(ACPI_LID);
+
+    /* no acpi lid object found */
+    if (lid_dir == NULL)
+	goto out;
+
+    while (1) {
+	lid_dent = readdir(lid_dir);
+	if (lid_dent == NULL) {
+	    /* no LID object */
+	    closedir(lid_dir);
+	    goto out;
+	}
+	if (strcmp(lid_dent->d_name, ".") &&
+		strcmp(lid_dent->d_name, "..")) {
+	    break;
+	}
+    }
+    state_name = malloc(strlen(ACPI_LID) + strlen(lid_dent->d_name) + 7);
+    memset(state_name, 0, sizeof(state_name));
+    strcat(state_name, ACPI_LID);
+    strcat(state_name, lid_dent->d_name);
+    strcat(state_name, "/state");
+
+    closedir(lid_dir);
+
+    if ((fd = open(state_name, O_RDONLY)) == -1) {
+	free(state_name);
+	goto out;
+    }
+    free(state_name);
+    if (read(fd, state, 64) == -1) {
+	close(fd);
+	goto out;
+    }
+    close(fd);
+    if (strstr(state, "open"))
+	ret = LID_OPEN;
+    else if (strstr(state, "closed"))
+	ret = LID_CLOSE;
+    else /* "unsupported" */
+	ret = LID_UNKNOWN;
+
+out:
+    if (pI830->debug_modes && (ret != LID_UNKNOWN))
+	xf86DrvMsg(pScrn->scrnIndex, X_INFO,
+		"LID switch detect %s with ACPI button\n",
+		ret ? "closed" : "open");
+
+    return ret;
+}
+
+/**
+ * Get LID switch close state from SWF
+ */
+static Bool
+i830_lvds_swf_lid_close(xf86OutputPtr output)
+{
+    ScrnInfoPtr pScrn = output->scrn;
+    I830Ptr pI830 = I830PTR(pScrn);
+    uint32_t swf14 = INREG(SWF14);
+    Bool ret;
+
+    if (swf14 & SWF14_LID_SWITCH_EN)
+	ret = TRUE;
+    else
+	ret = FALSE;
+
+    if (pI830->debug_modes)
+	xf86DrvMsg(pScrn->scrnIndex, X_INFO,
+		"LID switch detect %s with SWF14 0x%8x\n",
+		ret ? "closed" : "open", swf14);
+
+    return ret;
+}
+
 /**
  * Sets the power state for the panel.
  */
@@ -766,13 +877,21 @@ i830_lvds_mode_set(xf86OutputPtr output, DisplayModePtr mode,
 
 /**
  * Detect the LVDS connection.
- *
- * This always returns OUTPUT_STATUS_CONNECTED.  This output should only have
- * been set up if the LVDS was actually connected anyway.
  */
 static xf86OutputStatus
 i830_lvds_detect(xf86OutputPtr output)
 {
+    enum lid_status lid;
+
+    lid = i830_lvds_acpi_lid_open(output);
+    if (lid == LID_OPEN)
+	return XF86OutputStatusConnected;
+    else if (lid == LID_CLOSE)
+	return XF86OutputStatusDisconnected;
+
+    if (i830_lvds_swf_lid_close(output))
+	return XF86OutputStatusDisconnected;
+
     return XF86OutputStatusConnected;
 }
 
@@ -798,6 +917,11 @@ static void fill_detailed_block(struct detailed_monitor_section *det_mon,
         timing->misc |= 0x01;
 }
 
+/* X Server pre-1.5 compatibility */
+#ifndef DS_VENDOR
+#define DS_VENDOR 0x101
+#endif
+
 /**
  * Return the list of DDC modes if available, or the BIOS fixed mode otherwise.
  */
@@ -935,7 +1059,7 @@ static Atom panel_fitting_name_atoms[NUM_PANEL_FITTING_TYPES];
 
 
 static int
-i830_backlight_control_lookup(char *name)
+i830_backlight_control_lookup(const char *name)
 {
     int i;
 
@@ -992,7 +1116,7 @@ i830_lvds_set_backlight_control(xf86OutputPtr output)
 }
 
 static int
-i830_panel_fitting_lookup(char *name)
+i830_panel_fitting_lookup(const char *name)
 {
     int i;
 
@@ -1134,7 +1258,7 @@ i830_lvds_set_property(xf86OutputPtr output, Atom property,
     } else if (property == backlight_control_atom) {
 	INT32		    	backlight_range[2];
 	Atom			atom;
-	char			*name;
+	const char		*name;
 	int			ret, data;
 
 	if (value->type != XA_ATOM || value->format != 32 || value->size != 1)
@@ -1176,7 +1300,7 @@ i830_lvds_set_property(xf86OutputPtr output, Atom property,
 	return TRUE;
     } else if (property == panel_fitting_atom) {
 	Atom			atom;
-	char			*name;
+	const char		*name;
 	int			ret;
 
 	if (value->type != XA_ATOM || value->format != 32 || value->size != 1)
@@ -1291,6 +1415,13 @@ i830_lvds_init(ScrnInfoPtr pScrn)
     DisplayModePtr	    lvds_ddc_mode = NULL;
     struct i830_lvds_priv   *dev_priv;
 
+    if (!pI830->integrated_lvds) {
+	if (pI830->debug_modes)
+	    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
+		   "Skipping LVDS from driver feature BDB's LVDS config info.\n");
+	return;
+    }
+
     if (pI830->quirk_flag & QUIRK_IGNORE_LVDS)
 	return;
 
